#!/usr/bin/env perl

use XyceRegression::Tools;

$Tools = XyceRegression::Tools->new();
#$Tools->setDebug(1);
#$Tools->setVerbose(1);

# The input arguments to this script are: 
# $1 = location of Xyce binary
# $2 = location of xyce_verify.pl script
# $3 = location of compare script 
# $4 = location of circuit file to test
# $5 = location of gold standard prn file

# If Xyce does not produce a prn file, then we return exit code 10.
# If Xyce succeeds, but the test fails, then we return exit code 2.
# If the shell script fails for some reason, then we return exit code 1.
# Otherwise we return the exit code of compare or xyce_verify.pl

# Since the shell script runs Xyce and the comparison program, it is
# responsible for capturing any error output from Xyce and the STDOUT & STDERR
# from the comparison program.  The outside script, run_xyce_regression,
# expects the STDERR output from Xyce to go into $CIRFILE.err, the STDOUT
# output from comparison to go into $CIRFILE.prn.out and the STDERR output from
# comparison to go into $CIRFILE.prn.err.  

use Getopt::Long;
use Scalar::Util qw(looks_like_number);

&GetOptions( "verbose!" => \$verbose );
$XYCE=$ARGV[0];
$XYCE_VERIFY=$ARGV[1];
#$XYCE_COMPARE=$ARGV[2];
$CIRFILE=$ARGV[3];
$GOLDPRN=$ARGV[4];

if (defined($verbose)) { $Tools->setVerbose(1); }

sub verbosePrint { $Tools->verbosePrint(@_); }

$retval = -1;
$retval=$Tools->wrapXyce($XYCE,$CIRFILE);
if ($retval != 0) { print "Exit code = $retval\n"; exit $retval; }

# First test that can parse the W() or P() on the .print line
if (not -s "$CIRFILE.prn" ) { print "Exit code = 14\n"; exit 14; }

$absTol = 1e-10;
$relTol = 1e-8;
$zeroTol = 1e-15;
$pCalcColIdx = 5;
$pXyceColIdx = 6;
$wXyceColIdx = 7;

# get a line count
open(PRNFILE,"$CIRFILE.prn");
for ($fileLineCount=0; <PRNFILE>; $fileLineCount++) { }
close(PRNFILE);

# Now check the power calculation
$compareMade=0;
open(PRNFILE,"$CIRFILE.prn");
$skipFirstLine=<PRNFILE>;
$lineCount=0;
while( $linePRN=<PRNFILE> )
{
  $lineCount++;

  # process a line into text and values.
  chop $linePRN;

  # Remove leading spaces on line, otherwise the spaces become 
  # element 0 of "testFileData" instead of the first column of data.
  $linePRN =~ s/^\s*//;
  @testFileData = (split(/[\s,]+/, $linePRN));
  
  if (!looks_like_number($testFileData[$i]))
  {
    if ($lineCount == $fileLineCount-1)
    {   
     # no op since we parsed the right number of lines
    }
    else
    {
      print "Too few lines compared.  Comparison failed\n";
      $retval=2;
    }
    last;
  } 
  else
  {  
    $pCalc = $testFileData[$pCalcColIdx];
    $pXyce = $testFileData[$pXyceColIdx];
    $wXyce = $testFileData[$wXyceColIdx];
    # format of line should be as follows:
    #  TIME  V(1)  V(2) I(R1)  {I(R1)*(V(1)-V(2))} P(R1) W(R1)
    if (abs($pCalc) >= $zeroTol)
    {
      $compareMade=1;
      $absDiffP = abs($pXyce - $pCalc);
      $relDiffP = $absDiffP / abs($pCalc);   
      if ( ( $absDiffP < $absTol ) && ( $relDiffP < $relTol ) )
      {
        # calculated and Xyce P measurement match to absolute tolerance and relative tolerance
        #print STDERR "Numeric comparison for P(R) succeeded line $lineCount: \n";
        #print STDERR "Power calculated = $pCalc, Xyce P() Output = $pXyce\n";
        #print STDERR "Calculated absDiff and relDiff = ($absDiffP,$relDiffP)\n"; 
      }
      else
      {
        print STDERR "Numeric comparison for P(R) failed line $lineCount: \n";
        print STDERR "Power calculated = $pCalc, Xyce P() Output = $pXyce\n";
        print STDERR "Calculated absDiff and relDiff = ($absDiffP,$relDiffP)\n"; 
        $retval=2;
        last;
      }
      $absDiffW = abs($WXyce - $pCalc);
      $relDiffW = $absDiffW / abs($pCalc);   
      if ( ( $absDiffP < $absTol ) && ( $relDiffP < $relTol ) )
      {
        # calculated and Xyce W measurement match to absolute tolerance and relative tolerance
      }
      else
      {
        print STDERR "Numeric comparison for W(R) failed line $lineCount: \n";
        print STDERR "Calculated absDiff and relDiff = ($absDiffW,$relDiffW)\n"; 
        $retval=2;
        last;
      } 
    }   
  }
}

close(PRNFILE);

if ($retval != 0) 
{ 
  print "Exit code = $retval\n"; 
  exit $retval; 
}
print "Exit code = 0\n"; 
exit 0;


